@@ -1,6 +1,6 @@
 #!/usr/bin/perl -w
 
-# $Id: oinkgui.pl,v 1.42 2004/05/08 18:27:33 andreas_o Exp $ #
+# $Id: oinkgui.pl,v 1.49 2004/09/18 10:59:51 andreas_o Exp $ #
 
 # Copyright (c) 2004 Andreas stling <andreaso@it.su.se>
 # All rights reserved.
@@ -53,7 +53,6 @@ use constant CSIDL_DRIVES => 17;
 
 sub test_config();
 sub show_version();
-sub show_help();
 sub update_rules();
 sub clear_messages();
 sub create_cmdline($);
@@ -70,22 +69,18 @@ sub execute_oinkmaster(@);
 sub logmsg($ $);
 
 
-my $version = 'Oinkmaster GUI v1.0';
+my $version = 'Oinkmaster GUI v1.1';
 
 my @oinkmaster_conf = qw(
     /etc/oinkmaster.conf
     /usr/local/etc/oinkmaster.conf
 );
 
-# Graphical editors to look for.
-my @editors = qw(
-    kwrite kate kedit gedit xemacs xedit wordpad
-);
-
 # List of URLs that will show up in the URL BrowseEntry.
 my @urls = qw(
     http://www.snort.org/dl/rules/snortrules-snapshot-2_0.tar.gz
     http://www.snort.org/dl/rules/snortrules-snapshot-2_1.tar.gz
+    http://www.snort.org/dl/rules/snortrules-snapshot-2_2.tar.gz
     http://www.snort.org/dl/rules/snortrules-snapshot-CURRENT.tar.gz
 );
 
@@ -109,37 +104,42 @@ my %config = (
     careful          => 0,
     enable_all       => 0,
     check_removed    => 0,
-    mode             => 'normal',
+    output_mode      => 'normal',
+    diff_mode        => 'detailed',
+    perl             => $^X,
     oinkmaster       => "",
     oinkmaster_conf  => "",
     outdir           => "",
     url              => "",
     varfile          => "",
     backupdir        => "",
+    editor           => "",
 );
 
 my %help = (
 
   # File locations.
-    oinkscript   => 'Location of the executable Oinkmaster script (oinkmaster.pl).',
-    oinkconf     => 'The Oinkmaster configuration file to use.',
-    outdir       => 'Where to put the new rules. This should be the directory where you '.
-                    'store your current rules.',
+    oinkscript    => 'Location of the executable Oinkmaster script (oinkmaster.pl).',
+    oinkconf      => 'The Oinkmaster configuration file to use.',
+    outdir        => 'Where to put the new rules. This should be the directory where you '.
+                     'store your current rules.',
 
-    url          => 'Alternate location of rules archive to download/copy. '.
-                    'Leave empty to use the location set in oinkmaster.conf.',
-    varfile      => 'Variables that exist in downloaded snort.conf but not in '.
-                    'this file will be added to it. Leave empty to skip.',
-    backupdir    => 'Directory to put tarball of old rules before overwriting them. '.
-                    'Leave empty to skip backup.',
+    url           => 'Alternate location of rules archive to download/copy. '.
+                     'Leave empty to use the location set in oinkmaster.conf.',
+    varfile       => 'Variables that exist in downloaded snort.conf but not in '.
+                     'this file will be added to it. Leave empty to skip.',
+    backupdir     => 'Directory to put tarball of old rules before overwriting them. '.
+                     'Leave empty to skip backup.',
+    editor        => 'Full path to editor to execute when pressing the "edit" button '.
+                     '(wordpad is recommended on Windows). ',
 
   # Checkbuttons.
-    careful      => 'In careful mode, Oinkmaster will just check for changes, '.
-                    'not update anything.',
-    enable       => 'Some rules may be commented out by default (for a reason!). '.
-                    'This option will make Oinkmaster enable those.',
-    removed      => 'Check for rules files that exist in the output directory but not '.
-                    'in the downloaded rules archive.',
+    careful       => 'In careful mode, Oinkmaster will just check for changes, '.
+                     'not update anything.',
+    enable        => 'Some rules may be commented out by default (for a reason!). '.
+                     'This option will make Oinkmaster enable those.',
+    removed       => 'Check for rules files that exist in the output directory but not '.
+                     'in the downloaded rules archive.',
 
   # Action buttons.
     clear        => 'Clear current output messages.',
@@ -148,14 +148,11 @@ my %help = (
     update       => 'Execute Oinkmaster to update the rules.',
     test         => 'Test current Oinkmaster configuration. ' .
                     'If there are no fatal errors, you are ready to update the rules.',
-    help         => 'Execute oinkmaster -h.',
     version      => 'Request version information from Oinkmaster.',
 );
 
 
 my $gui_config_file = "";
-my $perl            = "";
-my $editor          = "";
 my $use_fileop      = 0;
 
 
@@ -192,26 +189,6 @@ foreach my $file (@oinkmaster_conf) {
     }
 }
 
-# Find out which editor to use.
-EDITOR:foreach my $ed (@editors) {
-    foreach my $dir (File::Spec->path()) {
-        my $file = "$dir/$ed";
-        if ((-f "$file" && -x "$file") || (-f "$file.exe" && -x "$file.exe")) {
-            $editor= $file;
-            last EDITOR;
-        }
-    }
-}
-
-# Look for Perl binary.
-PERL:foreach my $dir (File::Spec->path()) {
-    my $file = File::Spec->catfile("$dir", 'perl');
-    if ((-f "$file" && -x "$file") || (-f "$file.exe" && -x "$file.exe")) {
-        $perl = $file;
-        last PERL;
-    }
-}
-
 # Find out where the GUI config file is (it's not required).
 if ($ENV{HOME}) {
     $gui_config_file = "$ENV{HOME}/.oinkguirc"
@@ -341,6 +318,20 @@ my $backupdir_frame =
 $balloon->attach($backupdir_frame, -statusmsg => $help{backupdir});
 
 
+# Create frame with editor location.
+$filetypes = [
+  ['executable files', ['.exe']],
+  ['All files',    '*'                           ]
+];
+
+my $editor_frame =
+  create_fileSelectFrame($opt_tab, "Editor", 'EXECFILE',
+                         \$config{editor}, 'NOEDIT', $filetypes);
+
+$balloon->attach($editor_frame, -statusmsg => $help{editor});
+
+
+
 $notebook->pack(
   -expand => 'no',
   -fill   => 'x',
@@ -404,22 +395,35 @@ $balloon->attach(
 $left_frame->Label(
   -text        => "Output mode:",
   -background  => "$color{label}",
-)->pack(-side  => 'top',
-        -fill  => 'x',
+)->pack(
+  -side        => 'top',
+  -fill        => 'x',
 );
 
 # Create mode radiobuttons in the left frame.
-create_radiobutton($left_frame, "super-quiet", \$config{mode});
-create_radiobutton($left_frame, "quiet",       \$config{mode});
-create_radiobutton($left_frame, "normal",      \$config{mode});
-create_radiobutton($left_frame, "verbose",     \$config{mode});
+create_radiobutton($left_frame, "super-quiet", \$config{output_mode});
+create_radiobutton($left_frame, "quiet",       \$config{output_mode});
+create_radiobutton($left_frame, "normal",      \$config{output_mode});
+create_radiobutton($left_frame, "verbose",     \$config{output_mode});
 
+# Create "Diff mode" label.
+$left_frame->Label(
+  -text        => "Diff  mode:",
+  -background  => "$color{label}",
+)->pack(
+  -side  => 'top',
+  -fill  => 'x',
+);
+
+create_radiobutton($left_frame, "detailed",      \$config{diff_mode});
+create_radiobutton($left_frame, "summarized",    \$config{diff_mode});
+create_radiobutton($left_frame, "remove common", \$config{diff_mode});
 
 
 # Create "activity messages" label.
 $main->Label(
   -text       => "Output messages:",
-  -width      => '110',
+  -width      => '130',
   -background => "$color{label}",
 )->pack(
   -side       => 'top',
@@ -437,7 +441,7 @@ $out_frame->pack(
 
 # Pack help label below output window.
 $help_label->pack(
-    -fill       => 'x',
+    -fill     => 'x',
 );
 
 
@@ -458,11 +462,6 @@ $balloon->attach(
   -statusmsg => $help{version}
 );
 
-$balloon->attach(
-  create_actionbutton($left_frame, "Show help", \&show_help),
-  -statusmsg => $help{help}
-);
-
 $balloon->attach(
   create_actionbutton($left_frame, "Test configuration", \&test_config),
   -statusmsg => $help{test}
@@ -778,16 +777,15 @@ sub create_fileSelectFrame($ $ $ $ $ $)
                                      return;
                                  }
 
-                                 if ($editor) {
+                                 if ($config{editor}) {
                                      $main->Busy(-recurse => 1);
-                                     logmsg("Launching $editor. ".
-                                            "Close it to continue the GUI.\n\n", 'MISC');
+                                     logmsg("Launching " . $config{editor} .
+                                            ", close it to continue the GUI.\n\n", 'MISC');
                                      sleep(2);
-                                     system($editor, $$var_ref); # MainLoop will be put on hold...
+                                     system($config{editor}, $$var_ref); # MainLoop will be put on hold...
                                      $main->Unbusy;
                                  } else {
-                                     logmsg("No suitable editor found (looked for @editors ".
-                                            "in path)\n\n", 'ERROR');
+                                     logmsg("No editor set\n\n", 'ERROR');
                                  }
                              }
         )->pack(
@@ -835,18 +833,13 @@ sub show_version()
     $config{oinkmaster} =~ s/^\s+//;
     $config{oinkmaster} =~ s/\s+$//;
 
-    unless ($perl) {
-        logmsg("Perl binary not found in PATH!\n\n", 'ERROR');
-        return;
-    }
-
     unless ($config{oinkmaster} && -f "$config{oinkmaster}" &&
      (-x "$config{oinkmaster}" || $^O eq 'MSWin32')) {
         logmsg("Location of oinkmaster.pl is not set correctly!\n\n", 'ERROR');
         return;
     }
 
-    my $cmd = "$perl $config{oinkmaster} -V";
+    my $cmd = "$config{perl} $config{oinkmaster} -V";
     logmsg("$cmd:\n", 'EXEC');
     $main->Busy(-recurse => 1);
     my $output = `$cmd 2>&1` || "Could not execute $config{oinkmaster}: $!\n";
@@ -857,32 +850,6 @@ sub show_version()
 
 
 
-sub show_help()
-{
-    $config{oinkmaster} =~ s/^\s+//;
-    $config{oinkmaster} =~ s/\s+$//;
-
-    unless ($perl) {
-        logmsg("Perl binary not found in PATH!\n\n", 'ERROR');
-        return;
-    }
-
-    unless ($config{oinkmaster} && -f "$config{oinkmaster}" &&
-     (-x "$config{oinkmaster}" || $^O eq 'MSWin32')) {
-        logmsg("Location of oinkmaster.pl is not set correctly!\n\n", 'ERROR');
-        return;
-    }
-
-    my $cmd = "$perl $config{oinkmaster} -h";
-    logmsg("$cmd:\n", 'EXEC');
-    $main->Busy(-recurse => 1);
-    my $output = `$cmd 2>&1` || "Could not execute $config{oinkmaster}: $!\n";
-    $main->Unbusy;
-    logmsg("$output\n", 'OUTPUT');
-}
-
-
-
 sub execute_oinkmaster(@)
 {
     my @cmd = @_;
@@ -929,11 +896,6 @@ sub test_config()
 {
     my @cmd;
 
-    unless ($perl) {
-        logmsg("Perl binary not found in PATH!\n\n", 'ERROR');
-        return;
-    }
-
     create_cmdline(\@cmd) || return;
 
     push(@cmd, "-T");
@@ -1023,11 +985,6 @@ sub create_cmdline($)
         $$var_ref =~ s/\s+$//;
     }
 
-    unless ($perl) {
-        logmsg("Perl binary not found in PATH!\n\n", 'ERROR');
-        return (0);
-    }
-
     unless ($config{oinkmaster} && -f "$config{oinkmaster}" &&
      (-x "$config{oinkmaster}" || $^O eq 'MSWin32')) {
         logmsg("Location of oinkmaster.pl is not set correctly!\n\n", 'ERROR');
@@ -1053,16 +1010,18 @@ sub create_cmdline($)
     }
 
     push(@$cmd_ref,
-      "$perl", "$oinkmaster",
+      "$config{perl}", "$oinkmaster",
       "-C", "$oinkmaster_conf",
       "-o", "$outdir");
 
     push(@$cmd_ref, "-c")               if ($config{careful});
     push(@$cmd_ref, "-e")               if ($config{enable_all});
     push(@$cmd_ref, "-r")               if ($config{check_removed});
-    push(@$cmd_ref, "-q")               if ($config{mode} eq "quiet");
-    push(@$cmd_ref, "-Q")               if ($config{mode} eq "super-quiet");
-    push(@$cmd_ref, "-v")               if ($config{mode} eq "verbose");
+    push(@$cmd_ref, "-q")               if ($config{output_mode} eq "quiet");
+    push(@$cmd_ref, "-Q")               if ($config{output_mode} eq "super-quiet");
+    push(@$cmd_ref, "-v")               if ($config{output_mode} eq "verbose");
+    push(@$cmd_ref, "-m")               if ($config{diff_mode}   eq "remove common");
+    push(@$cmd_ref, "-s")               if ($config{diff_mode}   eq "summarized");
     push(@$cmd_ref, "-U", "$varfile")   if ($varfile);
     push(@$cmd_ref, "-b", "$backupdir") if ($backupdir);
 
