@@ -1,6 +1,6 @@
 #!/usr/bin/perl -w
 
-# $Id: oinkmaster.pl,v 1.331 2004/05/08 18:23:23 andreas_o Exp $ #
+# $Id: oinkmaster.pl,v 1.359 2004/10/01 18:30:26 andreas_o Exp $ #
 
 # Copyright (c) 2001-2004 Andreas stling <andreaso@it.su.se>
 # All rights reserved.
@@ -54,11 +54,13 @@ sub read_config($ $);
 sub sanity_check();
 sub download_file($ $);
 sub unpack_rules_archive($ $);
-sub process_rules($ $ $ $);
+sub process_rules($ $ $ $ $ $);
+sub process_rule($ $ $ $ $ $ $);
 sub setup_rules_hash($ $);
 sub get_first_only($ $ $);
 sub print_changes($ $);
 sub print_changetype($ $ $ $);
+sub print_summary_change($ $);
 sub make_backup($ $);
 sub get_changes($ $ $);
 sub get_new_filenames($ $);
@@ -73,11 +75,12 @@ sub parse_mod_expr($ $ $ $);
 sub untaint_path($);
 sub approve_changes();
 sub parse_singleline_rule($ $ $);
+sub minimize_diff($ $);
 sub catch_sigint();
 sub clean_exit($);
 
 
-my $VERSION            = 'Oinkmaster v1.0 by Andreas stling <andreaso@it.su.se>';
+my $VERSION            = 'Oinkmaster v1.1 by Andreas stling <andreaso@it.su.se>';
 my $OUTFILE            = 'snortrules.tar.gz';
 my $RULES_DIR          = 'rules';
 my $DIST_SNORT_CONF    = "$RULES_DIR/snort.conf";
@@ -93,9 +96,11 @@ my %config = (
     enable_all         => 0,
     interactive        => 0,
     make_backup        => 0,
+    minimize_diff      => 0,
     min_files          => 1,
     min_rules          => 1,
     quiet              => 0,
+    summary_output     => 0,
     super_quiet        => 0,
     update_vars        => 0,
     use_external_bins  => 1,
@@ -221,18 +226,23 @@ unpack_rules_archive("$tmpdir/$OUTFILE", $RULES_DIR);
 # Filenames (with full path) will be stored as %new_files{filename}.
 my $num_files = get_new_filenames(\my %new_files, "$tmpdir/$RULES_DIR");
 
-# Make sure we have at least the minumum number of files.
+# Make sure we have at least the minimum number of files.
 clean_exit("not enough rules files in downloaded archive (is it broken?)\n".
            "Number of rules files is $num_files but minimum is set to $config{min_files}.")
   if ($num_files < $config{min_files});
 
+# For possible localized rules.
+my %rh_tmp = setup_rules_hash(\%new_files, $config{output_dir});
+
 # Disable/modify/clean downloaded rules.
 my $num_rules = process_rules(\%{$config{sid_modify_list}},
                               \%{$config{sid_disable_list}},
                               \%{$config{sid_enable_list}},
+                              \%{$config{sid_local_list}},
+                              \%rh_tmp,
                               \%new_files);
 
-# Make sure we have at least the minumum number of rules.
+# Make sure we have at least the minimum number of rules.
 clean_exit("not enough rules in downloaded archive (is it broken?)\n".
            "Number of rules is $num_rules but minimum is set to $config{min_rules}.")
   if ($num_rules < $config{min_rules});
@@ -318,23 +328,25 @@ Usage: $progname -o <outdir> [options]
 This should be the directory where you store your Snort rules.
 
 Options:
--b <dir>   Backup your old rules into <dir> before overwriting them
--c         Careful mode - only check for changes and do not update anything
--C <cfg>   Use this configuration file instead of the default
-           May be specified multiple times to load multiple files
--e         Enable all rules that are disabled by default
--h         Show this usage information
--i         Interactive mode - you will be asked to approve the changes (if any)
--q         Quiet mode - no output unless changes were found
--Q         super-quiet mode (like -q but even more quiet when printing results)
--r         Check for rules files that exist in the output directory
-           but not in the downloaded rules archive
--T         Test configuration and then exit
--u <url>   Download from this URL instead of the one in the configuration file
-           (must be http://, https://, ftp://, file:// or scp:// ... .tar.gz)
--U <file>  Merge new variables from downloaded snort.conf into <file>
--v         Verbose mode
--V         Show version and exit
+-b <dir>  Backup your old rules into <dir> before overwriting them
+-c        Careful mode (dry run) - check for changes but do not update anything
+-C <cfg>  Use this configuration file instead of the default
+          May be specified multiple times to load multiple files
+-e        Enable all rules that are disabled by default
+-h        Show this usage information
+-i        Interactive mode - you will be asked to approve the changes (if any)
+-m        Minimize diff when printing result by removing common parts in rules
+-q        Quiet mode - no output unless changes were found
+-Q        super-quiet mode - like -q but even more quiet
+-r        Check for rules files that exist in the output directory
+          but not in the downloaded rules archive
+-s        Leave out details in rules results, just print SID, msg and filename
+-T        Test configuration and then exit
+-u <url>  Download from this URL instead of the one in the configuration file
+          (must be http://, https://, ftp://, file:// or scp:// ... .tar.gz)
+-U <file> Merge new variables from downloaded snort.conf into <file>
+-v        Verbose mode (debug)
+-V        Show version and exit
 
 RTFM
     exit;
@@ -356,10 +368,12 @@ sub parse_cmdline($)
         "e"   => \$$cfg_ref{enable_all},
         "h"   => \&show_usage,
         "i"   => \$$cfg_ref{interactive},
+        "m"   => \$$cfg_ref{minimize_diff},
         "o=s" => \$$cfg_ref{output_dir},
         "q"   => \$$cfg_ref{quiet},
         "Q"   => \$$cfg_ref{super_quiet},
         "r"   => \$$cfg_ref{check_removed},
+        "s"   => \$$cfg_ref{summary_output},
         "T"   => \$$cfg_ref{config_test_mode},
         "u=s" => \$$cfg_ref{url},
         "U=s" => \$$cfg_ref{varfile},
@@ -400,42 +414,114 @@ sub read_config($ $)
     my $config_file = shift;
     my $cfg_ref     = shift;
     my $linenum     = 0;
+    my $multi;
+    my %templates;
+
+    $config_file = File::Spec->canonpath(File::Spec->rel2abs($config_file));
 
     clean_exit("configuration file \"$config_file\" does not exist.\n")
       unless (-e "$config_file");
 
-    print STDERR "Loading " . File::Spec->rel2abs($config_file) . "\n"
+    print STDERR "Loading $config_file\n"
       unless ($config{quiet});
 
-    my ($dev, $ino) = (stat($config_file))[0,1]
-      or clean_exit("unable to stat $config_file: $!");
-
-  # Avoid loading the same file multiple times to avoid infinite recursion.
-    clean_exit("attempt to load \"$config_file\" twice.")
-      if ($loaded{$dev, $ino}++);
+  # Avoid loading the same file multiple times to avoid infinite recursion etc.
+    if ($^O eq "MSWin32") {
+        clean_exit("attempt to load \"$config_file\" twice.")
+          if ($loaded{$config_file}++);
+    } else {
+        my ($dev, $ino) = (stat($config_file))[0,1]
+          or clean_exit("unable to stat $config_file: $!");
+        clean_exit("attempt to load \"$config_file\" twice.")
+          if ($loaded{$dev, $ino}++);
+    }
 
     open(CONF, "<", "$config_file")
       or clean_exit("could not open configuration file \"$config_file\": $!");
     my @conf = <CONF>;
     close(CONF);
 
-    while ($_ = shift(@conf)) {
+    LINE:while ($_ = shift(@conf)) {
         $linenum++;
 
-      # Remove comments unless it's a modifysid line
-      # (the "#" may be part of the modifysid expression).
-        s/\s*\#.*// unless (/^\s*modifysid/i);
+      # Remove leading whitespaces and comment-only lines unless
+      # we're in the middle of reading a multi-line rule.
+        unless ($multi) {
+            s/^\s*//;
+            s/^#.*//;
+        }
 
-      # Remove leading/traling whitespaces.
-	s/^\s*//;
+      # Multi-line start/continuation?
+        if (/\\\s*\n$/) {
+            s/\\\s*\n$//;
+            $multi .= $_;
+            next LINE;
+        }
+
+      # Last line of multi-line directive?
+        if (defined($multi)) {
+            $multi .= $_;
+            $_ = $multi;
+            undef($multi);
+        }
+
+      # Remove traling whitespaces (*after* a possible multi-line is rebuilt).
 	s/\s*$//;
 
+      # Remove comments unless it's a modifysid/define_template line
+      # (the "#" may be part of the modifysid expression).
+        s/\s*\#.*// unless (/^modifysid/i || /^define_template/i);
+
       # Skip blank lines.
         next unless (/\S/);
 
+      # Use a template and make $_ a "modifysid" line.
+        if (/^use_template\s+(\S+)\s+([\d\s,\*]+)\s*(".*")*(?:#.*)*/i) {
+            my ($template_name, $sid, $args) = ($1, $2, $3);
 
-       # modifysid <SID[,SID, ...]> "substthis" | "withthis"
-       if (/^modifysids*\s+(\d+.*|\*)\s+"(.+)"\s+\|\s+"(.*)"\s*(?:#.*)*$/i) {
+            if (exists($templates{$template_name})) {
+               my $template = $templates{$template_name};  # so we don't substitute %ARGx% globally
+
+              # Evaluate each "%ARGx%" in the template to the corresponding value.
+                if (defined($args)) {
+                    my @args = split(/"\s+"/, $args);
+                    foreach my $i (1 .. @args) {
+                        $args[$i - 1] =~ s/^"//;
+                        $args[$i - 1] =~ s/"$//;
+                        $template =~ s/%ARG$i%/$args[$i - 1]/g;
+                    }
+                }
+
+              # There should be no %ARGx% stuff left now.
+                if ($template =~ /%ARG\d%/) {
+                    warn("WARNING: too few arguments for template \"$template_name\"\n");
+                    $_ = "error";  # so it will be reported as an invalid line later
+                }
+
+                unless ($_ eq "error") {
+                    $_ = "modifysid $sid $template\n";
+                    print STDERR "Template \"$template_name\" expanded to: $_"
+                      if ($config{verbose});
+                }
+
+            } else {
+                warn("WARNING: template \"$template_name\" has not been defined\n");
+            }
+        }
+
+      # new template definition.
+        if (/^define_template\s+(\S+)\s+(".+"\s+\|\s+".*")\s*(?:#.*)*$/i) {
+            my ($template_name, $template) = ($1, $2);
+
+            if (exists($templates{$template_name})) {
+                warn("WARNING: line $linenum in $config_file: ".
+                     "template \"$template_name\" already defined, keeping old\n");
+            } else {
+                $templates{$template_name} = $template;
+            }
+
+      # modifysid <SID[,SID, ...]> "substthis" | "withthis"
+        } elsif (/^modifysids*\s+(\d+.*|\*)\s+"(.+)"\s+\|\s+"(.*)"\s*(?:#.*)*$/i) {
             my ($sid_list, $subst, $repl) = ($1, $2, $3);
             warn("WARNING: line $linenum in $config_file is invalid, ignoring\n")
               unless(parse_mod_expr(\%{$$cfg_ref{sid_modify_list}},
@@ -452,6 +538,17 @@ sub read_config($ $)
 	        }
 	    }
 
+      # localsid <SID[,SID, ...]>
+        } elsif (/^localsids*\s+(\d.*)/i) {
+	    my $sid_list = $1;
+	    foreach my $sid (split(/\s*,\s*/, $sid_list)) {
+  	        if ($sid =~ /^\d+$/) {
+                    $$cfg_ref{sid_local_list}{$sid}++;
+	        } else {
+                    warn("WARNING: line $linenum in $config_file is invalid, ignoring\n");
+	        }
+	    }
+
       # enablesid <SID[,SID, ...]>
         } elsif (/^enablesids*\s+(\d.*)/i) {
 	    my $sid_list = $1;
@@ -527,10 +624,14 @@ sub sanity_check()
    my @req_params   = qw(path update_files);  # required parameters in conf
    my @req_binaries = qw(gzip tar);           # required binaries (unless we use modules)
 
-  # Can't use both -q and -v.
+  # Can't use both quiet mode and verbose mode.
     clean_exit("quiet mode and verbose mode at the same time doesn't make sense.")
       if ($config{quiet} && $config{verbose});
 
+  # Can't use multiple output modes.
+    clean_exit("can't use multiple output modes at the same time.")
+      if ($config{minimize_diff} && $config{summary_output});
+
   # Make sure all required variables are defined in the config file.
     foreach my $param (@req_params) {
         clean_exit("the required parameter \"$param\" is not defined in configuration file.")
@@ -718,8 +819,8 @@ sub download_file($ $)
 
         my $ua = LWP::UserAgent->new();
         $ua->env_proxy;
-        my $request = HTTP::Request->new(GET => $url);
-        my $response = $ua->request($request, $localfile);
+	my $request = HTTP::Request->new(GET => $url);
+	my $response = $ua->request($request, $localfile);
 
         clean_exit("could not download file: " . $response->status_line)
           unless $response->is_success;
@@ -819,7 +920,7 @@ sub unpack_rules_archive($ $)
         open(OLDOUT, ">&STDOUT")      or clean_exit("could not dup STDOUT: $!");
         open(STDOUT, ">$stdout_file") or clean_exit("could not redirect STDOUT: $!");
 
-        my $ret = system("tar", "tPf", "$archive");
+        my $ret = system("tar", "tf", "$archive");
 
         close(STDOUT);
         open(STDOUT, ">&OLDOUT") or clean_exit("could not dup STDOUT: $!");
@@ -900,11 +1001,13 @@ sub unpack_rules_archive($ $)
 # Open all rules files in the temporary directory and disable/modify all
 # rules/lines as requested in oinkmaster.conf, and then write back to the
 # same files. Also clean unwanted whitespaces and duplicate sids from them.
-sub process_rules($ $ $ $)
+sub process_rules($ $ $ $ $ $)
 {
     my $modify_sid_ref  = shift;
     my $disable_sid_ref = shift;
     my $enable_sid_ref  = shift;
+    my $local_sid_ref   = shift;
+    my $rh_tmp_ref      = shift;
     my $newfiles_ref    = shift;
     my %sids;
 
@@ -923,11 +1026,77 @@ sub process_rules($ $ $ $)
     print STDERR "\n"
       if ($config{verbose});
 
+  # Phase #1 - process all active rules and store in temporary hash.
+  # In case of dups, we use the one with the highest rev.
     foreach my $file (sort(keys(%$newfiles_ref))) {
 
-      # Make sure it's a regular file.
-        clean_exit("$file is not a regular file.")
-          unless (-f "$file" && !-l "$file");
+        open(INFILE, "<", "$file")
+          or clean_exit("could not open $file for reading: $!");
+	my @infile = <INFILE>;
+        close(INFILE);
+
+        my ($single, $multi, $nonrule, $msg, $sid);
+
+	RULELOOP:while (get_next_entry(\@infile, \$single, \$multi, \$nonrule, \$msg, \$sid)) {
+
+          # We don't care about non-rules in this phase.
+	    next RULELOOP if (defined($nonrule));
+
+          # Even if it was a single-line rule, we want a copy in $multi.
+	    $multi = $single unless (defined($multi));
+
+            my %rule = (
+                single => $single,
+                multi  => $multi,
+            );
+
+          # modify/disable/enable this rule as requested. Possible verbose 
+          # messages and warnings will be printed.
+            process_rule($modify_sid_ref, $disable_sid_ref, $enable_sid_ref, 
+                         \%rule, $sid, \%stats, 1);
+
+            $single = $rule{single};
+            $multi  = $rule{multi};
+
+          # Only care about active rules in this phase (the rule may have been
+          # disabled by a disablesid or a modifysid statement above, so we can't
+          # do this check earlier).
+	    next RULELOOP if ($multi =~ /^#/);
+
+          # Is it a dup? If so, see if this seems to be more recent (higher rev).
+            if (exists($sids{$sid})) {
+                warn("WARNING: duplicate SID in downloaded archive, SID=$sid\n")
+                  unless($config{super_quiet});
+
+                my ($old_rev) = ($sids{$sid}{single} =~ /\brev\s*:\s*(\d+)\s*;/);
+                my ($new_rev) = ($single             =~ /\brev\s*:\s*(\d+)\s*;/);
+
+              # This is so rules with a rev gets higher prio than 
+              # rules without any rev.
+                $old_rev = -1 unless (defined($old_rev));
+                $new_rev = -1 unless (defined($new_rev));
+
+              # If this rev is higher than the one in the last stored rule with
+              # this sid, replace rule with this one. This is also done if the
+              # revs are equal because we assume the rule appearing last in the
+              # rules file is the more recent rule.
+                if ($new_rev >= $old_rev) {
+                    $sids{$sid}{single} = $single;
+                    $sids{$sid}{multi}  = $multi;
+                }
+
+          # No dup.
+            } else {
+                $sids{$sid}{single} = $single;
+                $sids{$sid}{multi}  = $multi;
+            }
+        }
+    }
+
+  # Phase #2 - read all rules files again, but when writing active rules
+  # back to the files, use the one stored in the sid hash (which is free of dups).
+
+    foreach my $file (sort(keys(%$newfiles_ref))) {
 
         open(INFILE, "<", "$file")
           or clean_exit("could not open $file for reading: $!");
@@ -941,115 +1110,199 @@ sub process_rules($ $ $ $)
         my ($single, $multi, $nonrule, $msg, $sid);
 
 	RULELOOP:while (get_next_entry(\@infile, \$single, \$multi, \$nonrule, \$msg, \$sid)) {
-	    if (defined($nonrule)) {
-	        print OUTFILE "$nonrule";
-		next RULELOOP;
-	    }
-
-          # We've got a valid rule. If we have already seen this sid, discard this rule.
-            if (exists($sids{$sid})) {
-                warn("\nWARNING: duplicate SID in downloaded archive, SID $sid in ".
-                     basename($file) . " has already been seen in " .
-                     basename($sids{$sid}). ", discarding rule \"$msg\"\n");
+            if (defined($nonrule)) {
+                print OUTFILE "$nonrule";
                 next RULELOOP;
             }
 
-            $sids{$sid} = $file;
-
           # Even if it was a single-line rule, we want a copy in $multi.
-	    $multi = $single unless (defined($multi));
+            $multi = $single unless (defined($multi));
 
-          # Some rules may be commented out by default.
-          # Enable them if -e is specified.
-	    if ($multi =~ /^#/ && $config{enable_all}) {
-  	        print STDERR "Enabling disabled rule (SID $sid): $msg\n"
-	          if ($config{verbose});
-                $multi =~ s/^#*//;
-                $multi =~ s/\n#*/\n/g;
-	    }
+          # If this rule is marked as localized and has not yet been written,
+          # write the old version to the new rules file.
+            if (exists($$local_sid_ref{$sid}) && !exists($sids{$sid}{printed})) {
 
-          # Modify rule if requested.
-            my @all_mod = @{$$modify_sid_ref{'*'}}
-              if (exists($$modify_sid_ref{'*'}));
-
-            my @sid_mod = @{$$modify_sid_ref{$sid}}
-              if (exists($$modify_sid_ref{$sid}));
-
-            foreach my $mod_expr (@sid_mod, @all_mod) {
-
-                my ($subst, $repl) = ($mod_expr->[0], $mod_expr->[1]);
-		if ($multi =~ /$subst/s) {
-  	            print STDERR "Modifying SID $sid, subst=$subst, ".
-                                 "repl=$repl\nBefore: $multi\n"
-		      if ($config{verbose});
-
-                    $multi =~ s/$subst/$repl/see;
-
-  	  	    print STDERR "After:  $multi\n"
+              # Just ignore the rule in the downloaded file if it doesn't
+              # exist in the same local file.
+                unless(exists($$rh_tmp_ref{old}{rules}{basename($file)}{$sid})) {
+                    warn("WARNING: SID $sid is marked as local and exists in ".
+                         "downloaded " . basename($file) . " but the SID does not ".
+                         "exist in the local file, ignoring rule\n")
                       if ($config{verbose});
 
-                    $stats{modified}++;
-		} else {
-                    print STDERR "\nWARNING: SID $sid does not match ".
-                                 "modifysid expression \"$subst\", skipping\n"
-                      unless (exists($$modify_sid_ref{'*'}));
+                    next RULELOOP;
                 }
-	    }
 
-          # Disable rule if requested and it's not already disabled.
-            if (exists($$disable_sid_ref{$sid}) && $multi !~ /^\s*#/) {
-                print STDERR "Disabling SID $sid: $msg\n"
+                print OUTFILE $$rh_tmp_ref{old}{rules}{basename($file)}{$sid};
+                $sids{$sid}{printed} = 1;
+
+                warn("SID $sid is marked as local, keeping your version from ".
+                      basename($file) . ".\n".
+                     "Your version:       $$rh_tmp_ref{old}{rules}{basename($file)}{$sid}".
+                     "Downloaded version: $multi\n")
                   if ($config{verbose});
-                $multi = "#$multi";
-                $multi =~ s/\n([^#].+)/\n#$1/g;
-                $stats{disabled}++;
-	    }
 
-          # Enable rule if requested and it's not already enabled.
-            if (exists($$enable_sid_ref{$sid}) && $multi =~ /^\s*#/) {
-                print STDERR "Enabling SID $sid: $msg\n"
-                  if ($config{verbose});
-                $multi =~ s/^#+//;
-                $multi =~ s/\n#+(.+)/\n$1/g;
-                $stats{enabled}++;
-	    }
+                next RULELOOP;
+            }
 
-          # Write rule back to the same rules file.
-            print OUTFILE $multi;
+            my %rule = (
+                single => $single,
+                multi  => $multi,
+            );
+
+          # modify/disable/enable this rule. Possible verbose messages and warnings
+          # will not be printed (again) as this was done in the first phase.
+          # We send the stats to a dummy var as this was collected on the
+          # first phase as well.
+            process_rule($modify_sid_ref, $disable_sid_ref, $enable_sid_ref,
+                         \%rule, $sid, \my %unused_stats, 0);
+
+            $single = $rule{single};
+            $multi  = $rule{multi};
+
+          # Disabled rules are printed right back to the file, unless
+          # there also is an active rule with the same sid. Als o make
+          # sure we only print the sid once, even though it's disabled.
+            if ($multi =~ /^#/ && !exists($sids{$sid}) && !exists($sids{$sid}{printed})) {
+                print OUTFILE $multi;
+                $sids{$sid}{printed} = 1;
+                next RULELOOP;
+            }
+
+          # If this sid has not yet been printed and this is the place where
+          # the sid with the highest rev was, print the rule to the file.
+          # (There can be multiple totally different rules with the same sid
+          # and we don't want to put the wrong rule in the wrong place.
+            if (!exists($sids{$sid}{printed}) && $single eq $sids{$sid}{single}) {
+                print OUTFILE $multi;
+                $sids{$sid}{printed} = 1;
+            }
         }
 
         close(OUTFILE);
     }
 
     print STDERR "disabled $stats{disabled}, enabled $stats{enabled}, ".
-                 "modified $stats{modified}, total=" . keys(%sids) . ".\n"
+                 "modified $stats{modified}, total=$stats{total}\n"
       unless ($config{quiet});
 
-  # Warn on attempt at processing non-existent sids.
+  # Warn on attempt at processing non-existent sids, unless quiet mode.
     if ($config{verbose}) {
+
         foreach my $sid (keys(%$modify_sid_ref)) {
             next unless ($sid =~ /^\d+$/);    # don't warn on wildcard match
-            warn("WARNING: attempt to modify non-existent SID $sid\n")
+            warn("WARNING: attempt to use \"modifysid\" on non-existent SID $sid\n")
               unless (exists($sids{$sid}));
         }
+
         foreach my $sid (keys(%$enable_sid_ref)) {
-            warn("WARNING: attempt to enable non-existent SID $sid\n")
+            warn("WARNING: attempt to use \"enablesid\" on non-existent SID $sid\n")
               unless (exists($sids{$sid}));
         }
+
         foreach my $sid (keys(%$disable_sid_ref)) {
-            warn("WARNING: attempt to disable non-existent SID $sid\n")
+            warn("WARNING: attempt to use \"disablesid\" on non-existent SID $sid\n")
+              unless (exists($sids{$sid}));
+        }
+
+        foreach my $sid (keys(%$local_sid_ref)) {
+            warn("WARNING: attempt to use \"localsid\" on non-existent SID $sid\n")
               unless (exists($sids{$sid}));
         }
     }
 
   # Return total number of valid rules.
-    return (keys(%sids));
+    return ($stats{total});
+}
+
+
+
+# Process (modify/enable/disable) a rule as requested.
+sub process_rule($ $ $ $ $ $ $)
+{
+    my $modify_sid_ref  = shift;
+    my $disable_sid_ref = shift;
+    my $enable_sid_ref  = shift;
+    my $rule_ref        = shift;
+    my $sid             = shift;
+    my $stats_ref       = shift;
+    my $print_messages  = shift;
+
+  # Just for easier access.
+    my $single = $$rule_ref{single};
+    my $multi  = $$rule_ref{multi};
+
+    $$stats_ref{total}++;
+
+  # Some rules may be commented out by default.
+  # Enable them if -e is specified (both single-line and multi-line,
+  # version, because we don't know which version one we're going to
+  # use below.
+  # Enable them if -e is specified.
+    if ($multi =~ /^#/ && $config{enable_all}) {
+        $multi  =~ s/^#*//;
+        $multi  =~ s/\n#*/\n/g;
+        $single =~ s/^#*//;
+        $$stats_ref{enabled}++;
+    }
+
+  # Modify rule if requested. For disablesid/enablesid we work
+  # on the multi-line version of the rule (if exists). For
+  # modifysid that's no good since we don't know where in the
+  # rule the trailing backslashes and newlines are going to be
+  # and we don't want them to affect the regexp.
+    my @all_mod = @{$$modify_sid_ref{'*'}}
+      if (exists($$modify_sid_ref{'*'}));
+
+    my @sid_mod = @{$$modify_sid_ref{$sid}}
+      if (exists($$modify_sid_ref{$sid}));
+
+    foreach my $mod_expr (@sid_mod, @all_mod) {
+
+        my ($subst, $repl) = ($mod_expr->[0], $mod_expr->[1]);
+        if ($single =~ /$subst/si) {
+            print STDERR "Modifying SID $sid, subst=$subst, ".
+                         "repl=$repl\nBefore: $single\n"
+	      if ($print_messages && $config{verbose});
+
+          # Do the substitution on the single-line version and put it
+          # back in $multi.
+            $single =~ s/$subst/$repl/eei;
+            $multi = $single;
+
+  	    print STDERR "After:  $single\n"
+              if ($print_messages && $config{verbose});
+
+            $$stats_ref{modified}++;
+        } else {
+            print STDERR "WARNING: SID $sid does not match ".
+                         "modifysid expression \"$subst\", skipping\n"
+              if ($print_messages && !exists($$modify_sid_ref{'*'}));
+        }
+    }
+
+  # Disable rule if requested and it's not already disabled.
+    if (exists($$disable_sid_ref{$sid}) && $multi !~ /^\s*#/) {
+        $multi = "#$multi";
+        $multi =~ s/\n([^#].+)/\n#$1/g;
+        $$stats_ref{disabled}++;
+    }
+
+  # Enable rule if requested and it's not already enabled.
+    if (exists($$enable_sid_ref{$sid}) && $multi =~ /^\s*#/) {
+        $multi =~ s/^#+//;
+        $multi =~ s/\n#+(.+)/\n$1/g;
+        $$stats_ref{enabled}++;
+    }
+
+    $$rule_ref{single} = $single;
+    $$rule_ref{multi}  = $multi;
 }
 
 
 
 # Setup rules hash.
-# Format for rules will be:     rh{old|new}{rules{filename}{sid} = rule
+# Format for rules will be:     rh{old|new}{rules{filename}{sid} = single-line rule
 # Format for non-rules will be: rh{old|new}{other}{filename}     = array of lines
 # List of added files will be stored as rh{added_files}{filename}
 sub setup_rules_hash($ $)
@@ -1066,6 +1319,10 @@ sub setup_rules_hash($ $)
         warn("\nWARNING: downloaded rules file $file is empty\n")
           if (!-s "$file" && $config{verbose});
 
+      # Make sure it's a regular file.
+        clean_exit("\"$file\" is not a regular file.")
+          unless (-f "$file" && !-l "$file");
+
         open(NEWFILE, "<", "$file")
           or clean_exit("could not open $file for reading: $!");
         my @newfile = <NEWFILE>;
@@ -1094,7 +1351,7 @@ sub setup_rules_hash($ $)
 
 	    while (get_next_entry(\@oldfile, \$single, \$multi, \$nonrule, undef, \$sid)) {
 	        if (defined($single)) {
-		    warn("\nWARNING: duplicate SID in your local rules, SID ".
+		    warn("WARNING: duplicate SID in your local rules, SID ".
                          "$sid exists multiple times, please fix this manually!\n")
 		      if (exists($old_sids{$sid}));
 
@@ -1233,7 +1490,7 @@ sub make_backup($ $)
 
 
 
-# Print all changes.
+# Print the results.
 sub print_changes($ $)
 {
     my $ch_ref = shift;
@@ -1263,57 +1520,90 @@ sub print_changes($ $)
   # Print added rules.
     if (exists($$ch_ref{rules}{added})) {
         print "\n[+++]          Added rules:          [+++]\n";
-	print_changetype($PRINT_NEW, "Added to",
-                         \%{$$ch_ref{rules}{added}}, $rh_ref);
+        if ($config{summary_output}) {
+            print_summary_change(\%{$$ch_ref{rules}{added}}, $rh_ref);
+        } else {
+            print_changetype($PRINT_NEW, "Added to",
+                                 \%{$$ch_ref{rules}{added}}, $rh_ref);
+        }
     }
 
   # Print enabled rules.
     if (exists($$ch_ref{rules}{ena})) {
         print "\n[+++]         Enabled rules:         [+++]\n";
-	print_changetype($PRINT_NEW, "Enabled in",
-                         \%{$$ch_ref{rules}{ena}}, $rh_ref);
+        if ($config{summary_output}) {
+            print_summary_change(\%{$$ch_ref{rules}{ena}}, $rh_ref);
+        } else {
+            print_changetype($PRINT_NEW, "Enabled in",
+                             \%{$$ch_ref{rules}{ena}}, $rh_ref);
+        }
     }
 
   # Print enabled + modified rules.
     if (exists($$ch_ref{rules}{ena_mod})) {
         print "\n[+++]  Enabled and modified rules:   [+++]\n";
-	print_changetype($PRINT_BOTH, "Enabled and modified in",
-                         \%{$$ch_ref{rules}{ena_mod}}, $rh_ref);
+        if ($config{summary_output}) {
+            print_summary_change(\%{$$ch_ref{rules}{ena_mod}}, $rh_ref);
+        } else {
+   	    print_changetype($PRINT_BOTH, "Enabled and modified in",
+                             \%{$$ch_ref{rules}{ena_mod}}, $rh_ref);
+        }
     }
 
   # Print modified active rules.
     if (exists($$ch_ref{rules}{mod_act})) {
         print "\n[///]     Modified active rules:     [///]\n";
-	print_changetype($PRINT_BOTH, "Modified active in",
-                         \%{$$ch_ref{rules}{mod_act}}, $rh_ref);
+
+        if ($config{summary_output}) {
+            print_summary_change(\%{$$ch_ref{rules}{mod_act}}, $rh_ref);
+        } else {
+            print_changetype($PRINT_BOTH, "Modified active in",
+                             \%{$$ch_ref{rules}{mod_act}}, $rh_ref);
+        }
     }
 
   # Print modified inactive rules.
     if (exists($$ch_ref{rules}{mod_ina})) {
         print "\n[///]    Modified inactive rules:    [///]\n";
-	print_changetype($PRINT_BOTH, "Modified inactive in",
-                         \%{$$ch_ref{rules}{mod_ina}}, $rh_ref);
+        if ($config{summary_output}) {
+            print_summary_change(\%{$$ch_ref{rules}{mod_ina}}, $rh_ref);
+        } else {
+            print_changetype($PRINT_BOTH, "Modified inactive in",
+                             \%{$$ch_ref{rules}{mod_ina}}, $rh_ref);
+        }
     }
 
   # Print disabled + modified rules.
     if (exists($$ch_ref{rules}{dis_mod})) {
         print "\n[---]  Disabled and modified rules:  [---]\n";
-	print_changetype($PRINT_BOTH, "Disabled and modified in",
-                         \%{$$ch_ref{rules}{dis_mod}}, $rh_ref);
+        if ($config{summary_output}) {
+            print_summary_change(\%{$$ch_ref{rules}{dis_mod}}, $rh_ref);
+        } else {
+            print_changetype($PRINT_BOTH, "Disabled and modified in",
+                             \%{$$ch_ref{rules}{dis_mod}}, $rh_ref);
+        }
     }
 
   # Print disabled rules.
     if (exists($$ch_ref{rules}{dis})) {
         print "\n[---]         Disabled rules:        [---]\n";
-	print_changetype($PRINT_NEW, "Disabled in",
-                         \%{$$ch_ref{rules}{dis}}, $rh_ref);
+        if ($config{summary_output}) {
+            print_summary_change(\%{$$ch_ref{rules}{dis}}, $rh_ref);
+        } else {
+            print_changetype($PRINT_NEW, "Disabled in",
+                             \%{$$ch_ref{rules}{dis}}, $rh_ref);
+        }
     }
 
   # Print removed rules.
     if (exists($$ch_ref{rules}{removed})) {
         print "\n[---]         Removed rules:         [---]\n";
-	print_changetype($PRINT_OLD, "Removed from",
-                         \%{$$ch_ref{rules}{removed}}, $rh_ref);
+        if ($config{summary_output}) {
+            print_summary_change(\%{$$ch_ref{rules}{removed}}, $rh_ref);
+        } else {
+            print_changetype($PRINT_OLD, "Removed from",
+                             \%{$$ch_ref{rules}{removed}}, $rh_ref);
+        }
     }
 
 
@@ -1348,13 +1638,13 @@ sub print_changes($ $)
 
   # Print list of added files.
     if (keys(%{$$ch_ref{added_files}})) {
-        print "\n[+] Added files: [+]\n\n";
+        print "\n[+] Added files (consider updating your snort.conf to include them if needed): [+]\n\n";
         foreach my $added_file (sort({uc($a) cmp uc($b)} keys(%{$$ch_ref{added_files}}))) {
             print "    -> $added_file\n";
         }
     } else {
         print "\n[*] Added files: [*]\n    None.\n"
-          unless ($config{super_quiet});
+          unless ($config{super_quiet} || $config{summary_output});
     }
 
 
@@ -1363,13 +1653,13 @@ sub print_changes($ $)
     if ($config{check_removed}) {
         if (keys(%{$$ch_ref{removed_files}})) {
             print "\n[-] Files possibly removed from the archive ".
-                  "(consider removing them from your snort.conf): [-]\n\n";
+                  "(consider removing them from your snort.conf if needed): [-]\n\n";
             foreach my $removed_file (sort({uc($a) cmp uc($b)} keys(%{$$ch_ref{removed_files}}))) {
                 print "    -> $removed_file\n";
 	    }
         } else {
              print "\n[*] Files possibly removed from the archive: [*]\n    None.\n"
-               unless ($config{super_quiet});
+               unless ($config{super_quiet} || $config{summary_output});
         }
     }
 
@@ -1395,8 +1685,18 @@ sub print_changetype($ $ $ $)
             } elsif ($type == $PRINT_NEW) {
                 print "        $$rh_ref{new}{rules}{$file}{$sid}"
 	    } elsif ($type == $PRINT_BOTH) {
-                print "        old: $$rh_ref{old}{rules}{$file}{$sid}";
-                print "        new: $$rh_ref{new}{rules}{$file}{$sid}";
+
+                my $old = $$rh_ref{old}{rules}{$file}{$sid};
+                my $new = $$rh_ref{new}{rules}{$file}{$sid};
+
+                if ($config{minimize_diff}) {
+                    my ($old, $new) = minimize_diff($old, $new);
+                    print "\n        old SID $sid: $old";
+                    print "        new SID $sid: $new";
+                } else {
+                    print "\n        old: $old";
+                    print "        new: $new";
+                }
 	    }
         }
     }
@@ -1404,6 +1704,42 @@ sub print_changetype($ $ $ $)
 
 
 
+# Print changes in bmc style, i.e. only sid and msg, no full details.
+sub print_summary_change($ $)
+{
+    my $ch_ref = shift;   # reference to an entry in the rules changes hash
+    my $rh_ref = shift;   # reference to rules hash
+
+    my (@sids, %sidmap);
+
+    print "\n";
+
+  # First get all the sids (may be spread across multiple files.
+    foreach my $file (keys(%$ch_ref)) {
+        foreach my $sid (keys(%{$$ch_ref{$file}})) {
+            push(@sids, $sid);
+            if (exists($$rh_ref{new}{rules}{$file}{$sid})) {
+                $sidmap{$sid}{rule} = $$rh_ref{new}{rules}{$file}{$sid};
+            } else {
+                $sidmap{$sid}{rule} = $$rh_ref{old}{rules}{$file}{$sid};
+            }
+            $sidmap{$sid}{file} = $file;
+        }
+    }
+
+  # Print rules, sorted by sid.
+    foreach my $sid (sort {$a <=> $b} (@sids)) {
+        my @rule = $sidmap{$sid}{rule};
+        my $file = $sidmap{$sid}{file};
+	get_next_entry(\@rule, undef, undef, undef, \(my $msg), undef);
+        printf("%8d - %s ($file)\n", $sid, $msg);
+    }
+
+    print "\n";
+}
+
+
+
 # Compare the new rules to the old ones.
 sub get_changes($ $ $)
 {
@@ -1907,6 +2243,105 @@ sub approve_changes()
 
 
 
+# Remove common leading and trailing stuff from two rules.
+sub minimize_diff($ $)
+{
+    my $old_rule = shift;
+    my $new_rule = shift;
+
+    my $original_old = $old_rule;
+    my $original_new = $new_rule;
+
+  # Additional chars to print next to the diffing part.
+    my $additional_chars = 20;
+
+  # Remove the rev keyword from the rules, as it often
+  # makes the whole diff minimizing useless.
+    $old_rule =~ s/\s*\b(rev\s*:\s*\d+\s*;)\s*//;
+    my $old_rev = $1;
+
+    $new_rule =~ s/\s*\b(rev\s*:\s*\d+\s*;)\s*//;
+    my $new_rev = $1;
+
+  # If rev was the only thing that changed, we want to restore the rev
+  # before continuing so we don't remove common stuff from rules that
+  # are identical.
+    if ($old_rule eq $new_rule) {
+        $old_rule = $original_old;
+        $new_rule = $original_new;
+    }
+
+  # Temporarily remove possible leading # so it works nicely
+  # with modified rules that are also being either enabled or disabled.
+    my $old_is_disabled = 0;
+    my $new_is_disabled = 0;
+
+    $old_is_disabled = 1 if ($old_rule =~ s/^#//);
+    $new_is_disabled = 1 if ($new_rule =~ s/^#//);
+
+  # Go forward char by char until they aren't equeal.
+  # $i will bet set to the index where they diff.
+    my @old = split(//, $old_rule);
+    my @new = split(//, $new_rule);
+
+    my $i = 0;
+    while ($i <= $#old && $i <= $#new && $old[$i] eq $new[$i]) {
+        $i++;
+    }
+
+  # Now same thing but backwards.
+  # $j will bet set to the index where they diff.
+    @old = reverse(split(//, $old_rule));
+    @new = reverse(split(//, $new_rule));
+
+    my $j = 0;
+    while ($j <= $#old && $j <= $#new && $old[$j] eq $new[$j]) {
+        $j++;
+    }
+
+  # Print some additional chars on either side, if there is room for it.
+    $i -= $additional_chars;
+    $i = 0 if ($i < 0);
+
+    $j = -$j + $additional_chars;
+    $j = 0 if ($j > -1);
+
+    my ($old, $new);
+
+  # Print entire rules (i.e. they can not be shortened).
+    if (!$i && !$j) {
+        $old = $old_rule;
+        $new = $new_rule;
+
+  # Leading and trailing stuff can be removed.
+    } elsif ($i && $j) {
+        $old = "..." . substr($old_rule, $i, $j) . "...";
+        $new = "..." . substr($new_rule, $i, $j) . "...";
+
+  # Trailing stuff can be removed.
+    } elsif (!$i && $j) {
+        $old = substr($old_rule, $i, $j) . "...";
+        $new = substr($new_rule, $i, $j) . "...";
+
+  # Leading stuff can be removed.
+    } elsif ($i && !$j) {
+        $old = "..." . substr($old_rule, $i);
+        $new = "..." . substr($new_rule, $i);
+    }
+
+    chomp($old, $new);
+    $old .= "\n";
+    $new .= "\n";
+
+  # Restore possible leading # now.
+    $old = "#$old" if ($old_is_disabled);
+    $new = "#$new" if ($new_is_disabled);
+
+    return ($old, $new);
+}
+
+
+
 # Check a string and return 1 if it's a valid single-line snort rule.
 # Msg string is put in second arg, sid in third (those are the only
 # required keywords, besides the leading rule actions).
